/**************************************************************************
    MirrorFTP 

    Front-End to FTP.  Looks up hostname in mirrors database and gives a 
    list of sites which mirror files on the site specified.  If a host other
    than the one originally entered is selected then an ftp session is started
    at this address, otherwise it is started at the original host

    Bugs, comments, suggestions to dean@psy.uq.oz.au

    Acknowledgements: Thanks to Robert Dal Santo who came up with the 
                      original idea

 **************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef EXTENDED
#ifndef LOGGING
#define LOGGING
#endif
#define LOG_UID 1
#else
#define LOG_UID 0
#endif

#ifdef LOGGING
#include <syslog.h>
#endif

/* Constants */
#ifndef DBFILE
#define DBFILE "/etc/ftp-mirrors"   /* Database file which holds mirror info */
#endif

#ifndef REALFTP
#define REALFTP "/usr/bin/real-ftp" /* Where the real binary sits */
#endif

#define AUSDOMAIN "au"              /* Domain for australia */

#define HOSTSIZE 80          /* Max Size of a hostname */ 
#define DESCSIZE 80          /* Max size of a description */
#define MAXMIRRORS 20        /* Max number of mirror entries allowed */
#define LINESIZE 256         /* Max size of a line */
#define DIRSIZE 80           /* Max directory name size */
#define FS ':'               /* Field seperator in db */

/* Types */
typedef struct dirs
{
    char rDir[DIRSIZE];
    char lDir[DIRSIZE];
    char desc[DESCSIZE];
    struct dirs *next;
} DIRS;

typedef struct
{
    char mirror[HOSTSIZE];
    DIRS *directories;    
} MIRRORS;


/* prototypes */
static char *CheckForMirror();
static void StripComments();
static char *ChooseSite();
static int GetFields();
static int AddDir();
static int HostMatches();
static int IsDomain();
static int strcasecomp();

int main(argc,argv)
int argc;
char *argv[];
{
    int i=0;
    unsigned addr;
    char *mirror=0;
    char *oldhostname;
    struct hostent *hostInfo;

#ifdef LOGGING
    /* Set syslog facility */ 
    openlog("mirrorftp",0,LOGFACILITY);
#endif

    /* Search for the hostname (first non switch arg) */
    if (argc>1) for(i=1; i<argc && (*(argv[i]) == '-'); i++);

    /* Get Mirrors if there are any */
    if (i != 0 && i != argc)
    {
        /* Get the host info for this host */
        hostInfo = gethostbyname(argv[i]);
        if (!hostInfo) 
        {
             addr = inet_addr(argv[i]);
             hostInfo = gethostbyaddr((char *)&addr, 4, AF_INET);

             /* If null then display error */
             if (!hostInfo) 
             {
                 herror(argv[i]);
                 argv[i]=0; 
             }
        }
        else
        {

#ifdef LOGGING
            /* Save the hostname */
            oldhostname = strdup(argv[i]);
#endif
 
#ifdef NOAUSSITES
            /* Check to see if this domain ends in au, if so then ignore */
            if (!IsDomain(hostInfo->h_name,AUSDOMAIN))
#endif
            mirror = CheckForMirror(argv[i],hostInfo);
            if (mirror) argv[i] = mirror;
        }
    }

#ifdef LOGGING
    if (i && argv[i] && i != argc)
    {
        char msg[256];

        /* Log the statistics */
        if (LOG_UID)
        {
            if (mirror) sprintf(msg,"M:%s => %s:%d", oldhostname,
                                                     mirror,getuid());
            else sprintf(msg,"U:%s:%d", oldhostname, getuid());
        }
        else
        {
            if (mirror) sprintf(msg,"M:%s => %s", oldhostname, mirror);
            else sprintf(msg,"U:%s", oldhostname);
        }
        syslog(LOG_INFO,msg);
     }
#endif

    /* Set argv[0] to the real ftp command */
    argv[0]=REALFTP;

    /* Exec the real ftp with the modified args */
    if (execv(REALFTP,argv)==-1)
    {
        perror("Could not exec ftp"); 
    }

    return 0;
}


/**************************************************************************
    FUNCTION: static char *CheckForMirror(const char* host, 
                                          struct hostent *hostInfo )
    
    PURPOSE : Look up a host in the mirrors database and list all the 
              sites that mirror it.  Presents the user with a list of 
              options to choose from and returns the host they picked

***************************************************************************/ 
static char *CheckForMirror(host,hostInfo)
char *host;
struct hostent *hostInfo;
{
    /* Look up host in the database to see if it can be found */

    FILE *database;       /* The database file containing the mirror entries */ 
    int line=0;                    /* Number of lines processed */
    MIRRORS mirrorsites[MAXMIRRORS]; /* List of mirror sites */
    int nrMirrors=0; 

    /* Open database */
    if (!(database=fopen(DBFILE,"r")))
    { 
        perror("Cannot open ftp-mirrors file");
        return 0;
    }
    
    /* Sequential Search through the file for each mirror */
    while (!feof(database))
    {
        char in[LINESIZE];     /* String read in */
        char lHost[HOSTSIZE];  /* The host being mirrored */
        char ip[HOSTSIZE];     /* The IP Address of host being mirrored */
        char mirror[HOSTSIZE]; /* The name of the mirror */
        char desc[DESCSIZE];   /* Description of these files */
        char lDir[DIRSIZE];   /* The directory on the mirror */
        char rDir[DIRSIZE];   /* The original directory */

        /* Init values to 0 length */
        *in=0;
        *lHost=0;
        *mirror=0;
        *desc=0;
 
        /* Get the database fields */
        if (!fgets(in, LINESIZE, database)) continue;
        line++;

        /* strip comments */
        StripComments(in);
        if (!*in || *in=='\n') continue; /* Skip blanks */
         
        /* Get fields from database */
        if ((GetFields(in, lHost, ip, mirror, lDir, rDir, desc)==-1) || 
            (!*lHost ^ !*mirror ))
        {
            fprintf(stderr,"Error in config at line %d\n", line);
            exit(-1);
        }

        /* Check for a match */
        if (HostMatches(lHost,ip,hostInfo))
        {
            /* Add mirror to list */
            AddDir(mirrorsites,&nrMirrors,mirror,lDir,rDir,desc);
        }
 
        /* Check to make sure nrMirrors is not exceeded */
        if (nrMirrors == MAXMIRRORS) break;
    }

    /* If no matches then return null */
    if (!nrMirrors) return 0;
   
    /* Add the site passed to the end */ 
    strcpy(mirrorsites[nrMirrors].mirror,host);
    mirrorsites[nrMirrors].directories = 0;
    
    /* Get the site they want */
    return ChooseSite(mirrorsites,nrMirrors);
}

/**************************************************************************
    FUNCTION: static void StripComments(char *line) 
    
    PURPOSE : Removes comments from the line
**************************************************************************/
static void StripComments(line)
char *line;
{
    char *offset=0; 
 
    /* Get rid of comment */
    offset=strrchr(line,'#');
    if (offset) *offset = '\0';

    /* Remove trailing whitespace */
    while(isspace(*line)) line++;

    return; 
}

/**************************************************************************
    FUNCTION: static char *ChooseSite(MIRRORS m[], int nrSites) 
    
    PURPOSE : Displays a list of sites and returns the one the user picks 
**************************************************************************/
static char *ChooseSite(m,nrSites)
MIRRORS m[]; 
int nrSites;
{
    int i=0;
    char in[LINESIZE];
    int number=0;
    DIRS *temp;

    /* Display the header message */
    printf("\nThe following sites mirror %s.\n", 
            m[nrSites].mirror);
    puts("These hosts may or may not be up to date but should");
    puts("give considerably better performance");

    /* Display the sites */
    for (i=0; i<=nrSites; i++)
    {
        printf("\n%d. %s\n", i+1, m[i].mirror);
        
        /* Print each of the fields */
        if (m[i].directories)
        { 
            /* Print remote dir => local dir for each directory */
            temp = m[i].directories;
 
            while(temp)
            { 
 
                if (*temp->rDir) printf("   %s ", temp->rDir);
                if (*temp->lDir) printf("=> %s ",temp->lDir);

                /* Display description */
                if (*temp->desc) printf("(%s)",  temp->desc);
                putchar('\n');
                temp = temp->next;
            }
        }
    }

    /* Get the choice */
    do
    {
        /* Display prompt */
        printf("\nChoose the number of the site to connect to:");
         
        /* Get the input */
        gets(in);

        /* Check the input */
        if (*in == 'q') exit(0); /* Exit */
        if (feof(stdin))         /* Abort */
        {
            putchar('\n');
            exit(-1);
        }
        if (!*in) number = 1; /* Default */
        else number = atoi(in);
        if (number > nrSites+1 || number < 1) puts("** Invalid choice **");

    } while (number > nrSites+1 || number < 1);

   /* Return the corresponding host */
   return m[number-1].mirror;
} 

/**************************************************************************
    FUNCTION: static int GetFields(in,lhost,ip,mirror,lDir, rDir, desc)
    
    PURPOSE : Splits line on a colon into fields specified 
**************************************************************************/
static int GetFields(in,lHost,ip,mirror,lDir, rDir, desc)
char *in;
char *lHost;
char *ip;
char *mirror;
char *lDir;
char *rDir;
char *desc;

{
     int i=0;
     int length=0;
     int nrFields=0;
     char *line=in;

     /* Initialise to null strings */
     *lHost=0;
     *ip=0;
     *mirror=0;
     *lDir=0;
     *rDir=0;
     *desc=0;

     /* Get length of line */
     length=strlen(in);

     /* Split the string */
     for(i=0; i<length; i++) 
     {
         if(in[i] == FS) 
         {
             in[i]=0; 
             nrFields++;
         }
     }
   
     /* Check to see if the right number of fields was returned */
     if (nrFields+1 != 6) return -1;
   
     /* Get the values from the string */
     strcpy(lHost,line);
     line+=strlen(line)+1;
     if (line-in>=length) return -1;

     strcpy(ip,line);
     line+=strlen(line)+1;
     if (line-in>=length) return -1;

     strcpy(mirror,line);
     line+=strlen(line)+1;
     if (line-in>=length) return -1;

     strcpy(rDir, line);
     line+=strlen(line)+1;
     if (line-in>=length) return -1;

     strcpy(lDir, line);
     line+=strlen(line)+1;
     if (line-in>=length) return -1;

     strcpy(desc, line);
     if (desc[strlen(desc)-1]=='\n') 
     {
         desc[strlen(desc)-1]='\0';
     }
     else
     {
         return -1;
     }	
     
     return 0;
}

/**************************************************************************
    FUNCTION: static int AddDir(m,nrMirrors,mirror,lDir,rDir,desc)
    
    PURPOSE : Adds a mirrored directory to the list
**************************************************************************/
static int AddDir(m,nrMirrors,mirror,lDir,rDir,desc)
MIRRORS m[];
int *nrMirrors;
char *mirror;
char *lDir;
char *rDir;
char *desc;
{
    int i;
    DIRS *temp;
    static DIRS* entry;

    /* Make entry */
    entry = (DIRS *) malloc(sizeof(DIRS));
    strcpy(entry->lDir,lDir);
    strcpy(entry->rDir,rDir);
    strcpy(entry->desc,desc);
    entry->next = 0;

    /* Search list to see where to add directory.  This will either be at
    ** the end of the list, or where a dir for this mirror was already 
    ** added 
    */
    for (i=0; i<*nrMirrors && strcmp(m[i].mirror,mirror); i++);

    /* Add mirror if new */
    if (i==*nrMirrors)
    {
        /* Set the fields for this mirror */
        strcpy(m[i].mirror,mirror);
        m[i].directories=entry; 
        (*nrMirrors)++;
    }    
    else
    {
        /* Find the tail of the list and tack on this entry */
        temp = m[i].directories;
        while(temp->next) temp=temp->next;
        temp->next = entry;
    }  
    return 0;
}

/**************************************************************************
    FUNCTION: static int HostMatches(name,ip,hostInfo) 
    
    PURPOSE : Returns true or false depending on whther a name or ip 
              address are matched by any of the fields in the hostent  
**************************************************************************/
static int HostMatches(name,ip,hostInfo)
char *name; 
char *ip; 
struct hostent *hostInfo;
{
    char **temp=0;

    /* Try official name */
    if (strcasecomp(hostInfo->h_name,name)==0) return 1;

    /* Try IPs */
    if (hostInfo->h_addrtype != AF_INET) return 0;
    temp = hostInfo->h_addr_list;
    while (*temp)
    {
       struct in_addr *in = (struct in_addr *) *temp;
       if (strcasecomp(inet_ntoa(*in),ip)==0) return 1;
       else temp++;
    }

    /* Try aliases */
    temp=hostInfo->h_aliases;
    while(*temp)
    {
        if (strcasecomp(*temp,name)==0) return 1;     
        else temp++; 
    }
    return 0;
}

/**************************************************************************
    FUNCTION: int IsDomain(name,domain) 
    
    PURPOSE : Returns true or false if the name specified ends in the domain
              specified. 
**************************************************************************/
static int IsDomain(name,domain)
char *name; 
char *domain;
{
    /* Compare the last bit of the official host address */
    return (strcasecomp(name+(strlen(name)-strlen(domain)), domain)==0);
}

/***************************************************************************
    FUNCTION: int strcasecomp(s1,s2)
    
    PURPOSE:  Case insensitive Compare two strings  
****************************************************************************/
static int strcasecomp(s1,s2)
const char *s1; 
const char *s2;
{
    int i;
    int cmp;

    /* Loop through strings until they are not equal or the end of one is 
    ** reached */
    for(i=0;s1[i] && s2[i]; i++)
    {
        cmp=tolower(s1[i]) - tolower(s2[i]);
        if (cmp) return cmp;
    }

    /* If both the strings are equal length this will return 0 else it will
    ** return which one is greater */ 
    return s1[i]-s2[i];
}
